#ifndef BASE_CONSOLE_H
#define BASE_CONSOLE_H

#include <vector>
#include <sstream>
#include <iostream>
#include "math/vec.h"
#include "system/mutex.h"
#include "system/keyboard.h"
#include "texture.h"

class ConsoleInterpreter {
public:
	virtual void exec(const char *cmd) = 0;
};

class Console {
public:
	Console();
	~Console() {shutdown();}
	
	void init(int w, int h);
	void shutdown();

	//gathers all keystrokes recorded in the last frame - interprets them
	void preUpdate();

	//fixes whatever damage PreUpdate did to the input headed toward the application layer
	void postUpdate();

	//adds a string to the output window (called by print)
	void addString(const char *appendData);

	//shell for AddString() - interprets the *_format
	void print(const char *_format, ...);

	//displays the console
	void display();

	//sets whether the console is visible when inactive
	void setIdleVisible(bool isvis);

	//used via console command - set the console's info (drop time, width, height)
	void setInfo(int argc, char **argv);
	
	//modularity.  defaults to the cvar interpreter
	void setInterpreter(ConsoleInterpreter *interpreter);

private:

	//issued when Update() records a key being pressed
	bool keyDown(int key);

	//when keyDown() records a '~' - toggles console display & input trapping
	void toggle();

	//when keyDown() records anything else - add a key to the input buffer
	void addToInput(char key);

	//clear the current input buffer
	void clearInput();

	//displays a console line - called by Display()
	void displayLine(char *p, int len);

	int historyPos;							//used for cycling thru the prev cmd buffer
	int cursorPos;
	std::vector<std::string> history;
	char currentcmd[256];
	char textbuffer[8192];		//the current text buffer

	//no one extend the keyboard any time soon
	PKeyboardHandler *pushKeyboardHandler;
	bool usePushKeyboardHandler;
	
	ConsoleInterpreter *interpreter;

	//for locking the console
	Mutex mutex;

//this is me being bad.
//we really need a separate font system
public:
	Texture font;		//texture class of the font used
private:

	//used for toggling the console
	bool isActive;

//modifyable variables
	//used for whether the console is visible when idle
	bool idleVisible;

	//amount of chars to display in the console
	vec2i size;

public://no harm in on-the-fly modifications of this variable
	//how far to drop when idle
	int idleDropdown;
private:

	//how long it takes to drop
	double droptime;

	//the type of dropdown display method
	int droptype;
	int usedDropType;

//internally used variables
	//when the console starts to drop
	double dropstart;		//time_t

	//the percentage of the timestep the console is through the dropdown procedure
	float timefrac;
};

//////////////// my attempt at C++izing this all ////////////////

#include <string>
#include <sstream>
#include <ostream>

//if only Console were an ostream ... then it would inherit all of these ...
//maybe it should be ...
//until then, lets let all unknown types fall into this template
template<typename T> inline Console &operator<<(Console &con, const T &t) {
	std::ostringstream o;
	o << t;
	con.addString(o.str().c_str());
	return con;
}

//with char * and std::string, just write it directly to the text buffer (don't re-printf it)
inline Console &operator<<(Console &con, const char *str) { con.addString(str); return con; }
inline Console &operator<<(Console &con, const std::string &s) { con.addString(s.c_str()); return con; }
inline Console &operator<<(Console &con, float f) { con.print("%f", f); return con; }
inline Console &operator<<(Console &con, double f) { con.print("%f", f); return con; }
inline Console &operator<<(Console &con, int f) { con.print("%d", f); return con; }
inline Console &operator<<(Console &con, unsigned int f) { con.print("%u", f); return con; }

#ifndef WIN32	//looks like in win32 size_t is already defined ...
inline Console &operator<<(Console &con, size_t f) { con.print("%u", f); return con; }
#endif //WIN32

//endl support
inline Console &operator<<(Console &con, std::ostream &(*pf)(std::ostream &)) {
	std::ostream &(*el)(std::ostream &) = std::endl;
	if ((void*)pf == (void*)el) {
		con << "\n";
	}
	return con;
}

#endif	//BASE_CONSOLE_H
