#ifndef BASE_CVAR_H
#define BASE_CVAR_H

#include "defs/singleton.h"
#include "math/vec.h"
#include "console.h"	//make sure console.h doesn't include cvar.h.  it doesn't need to.  only console.cpp

#include <stdlib.h>	//atof on linux
#include <string>
#include <list>

extern Console con;	//kind of a mess ...

typedef void CFUNC(int, char **);

class CVar {
private:
	const char *name;
	const char *desc;
public:
	CVar(const char *name);
	CVar(const char *name, const char *desc);
	CVar(const char *name, CFUNC *func);
	CVar(const char *name, const char *desc, CFUNC *func);

	virtual void execArgs(int argc, char **argv);

	const char *getName() { return name; }

	CFUNC *func;
};

class CVarPrim : public CVar {
public:
	CVarPrim(const char *name, const char *desc, CFUNC *func) : CVar(name,desc,func) {}

	virtual void execArgs(int argc, char **argv);
	virtual void setValue(int argc, char **argv);
	virtual void echoData();
};

template<typename T>
class _cvar : public CVarPrim {
public:
	T data;

	_cvar(const char *name, const T &val) : CVarPrim(name,NULL,NULL), data(val) {}

	virtual void setValue(int argc, char **argv) { if (argc > 1) data = (T)atof(argv[1]); }
	virtual void echoData() { con << getName() << " := " << data << "\n"; }
	
	virtual T &operator=(const T &val) { data = val; return data; }
	virtual operator T() { return data; }
	virtual operator const T() const { return data; }
};

typedef _cvar<int> CVari;
typedef _cvar<float> CVarf;

template<int DIM, typename T>
class _cvar<_vec<DIM,T> > : public CVarPrim {
public:
	typedef _vec<DIM,T> vec_t;

	vec_t data;

	_cvar(const char *name, const vec_t &val)
	: CVarPrim(name,NULL,NULL), data(val) {}
	
	virtual void setValue(int argc, char **argv) {
		data.set((T)0);
		for (int i = 0; i < argc && i < data.dim(); i++) {
			if (argv[i]) data[i] = (T)atof(argv[i]);
		}
	}
	
	virtual void echoData() {
		con << getName() << " :=";
		for (int i = 0; i < data.dim(); i++) {
			con << " " << data[i];
		}
		con << "\n";
	}

	virtual void operator=(const vec_t &f) { data = f; }
	virtual operator vec_t() { return data; }
	virtual operator const vec_t() const { return data; }
};

typedef _cvar<vec2f> CVar2f;

typedef enum {
	CVAR_TYPE_NONE,
	CVAR_TYPE_BOOL,
	CVAR_TYPE_INT,
	CVAR_TYPE_FLOAT,
	CVAR_TYPE_DOUBLE,
} cvar_type_t;

//wildcard - auto-detect the primitive - but not operator= or typecast operations
class CVarw : public CVarPrim {
public:
	CVarw(int data, const char *name, CFUNC *func = NULL, int _dim = 0);
	CVarw(bool *data, const char *name, CFUNC *func = NULL, int _dim = 1);
	CVarw(int *data, const char *name, CFUNC *func = NULL, int _dim = 1);
	CVarw(float *data, const char *name, CFUNC *func = NULL, int _dim = 1);
	CVarw(double *data, const char *name, CFUNC *func = NULL, int _dim = 1);
	//math lib specific
	CVarw(vec3f *data, const char *name, CFUNC *func = NULL);
	CVarw(vec4f *data, const char *name, CFUNC *func = NULL);
	CVarw(quatf *data, const char *name, CFUNC *func = NULL);

	void *data;
	cvar_type_t type;
	int dim;

	virtual void setValue(int argc, char **argv);
	virtual void echoData();
};

#define ADD_CVARW(x)	static CVarw cv_##x(&x, #x)

//modularity for the console
class CVarInterpreter : public ConsoleInterpreter {
public:
	virtual void exec(const char *cmd);
};
extern CVarInterpreter cvarInterpreter;

class CVarSys {
public:
	void init();
	void shutdown();
	void exec(const char *cmd);

	std::list<CVar *> cvars;

	void addNew(CVar *v);
	CVar *matchString(const char *str);
	void parseArgs(const char *buffer, std::list<std::string> &args);
	void executeArgs(const std::list<std::string> &args);
};
extern Singleton<CVarSys> cvarSys;

#endif