#ifndef BASE_IMAGE_H
#define BASE_IMAGE_H

#include "math/vec.h"

/**
 * Image class used for texture loading.
 * This stuff should probably go in its own file.
 */
class Image {
public:
	//it would be nice to protect these someday...
	vec2i size;
	int bitsPerPixel;
	unsigned char *data;

	void reset();
	void unload();
	
	Image()	{ reset(); }

	Image(const vec2i &_size, int _bitsPerPixel);
protected:
	//this just takes the memory.  no copy constructor.  just TAKES it.
	Image(const vec2i &_size, int _bitsPerPixel, unsigned char *_data)
	: size(_size), bitsPerPixel(_bitsPerPixel), data(_data) {}
public:
	virtual ~Image() { unload(); }
	
	int getRowSize() const {
		int rowbits = size.x * bitsPerPixel;
		return (rowbits >> 3) + !!(rowbits & 7);
	}

	int getDataSize() const {
		return getRowSize() * size.y;
	}
};

//pixel iteration
template<typename ImagePtr, typename PixelRef, typename PixelPtr>
class ImagePixelIterator {
	ImagePtr src;
	vec2i mtc;
	int mindex;
public:

	ImagePixelIterator(ImagePtr src_, int index_)
	: src(src_), mindex(index_) {
		mtc.x = mindex % src->size.x;
		mtc.y = mindex / src->size.x;
	}
	ImagePixelIterator &operator++() {
		mindex++;
		mtc.x++;
		if (mtc.x >= src->size.x) {
			mtc.x -= src->size.x;
			mtc.y++;
		}
		return *this;
	}
	bool operator==(const ImagePixelIterator &other) { return mindex == other.mindex; }
	bool operator!=(const ImagePixelIterator &other) { return !operator==(other); }
	const vec2i &tc() const { return mtc; }
	int index() const { return mindex; }
	PixelPtr operator->() { return (PixelPtr)(src->data + ((mtc.x + mtc.y * src->size.x) * src->bitsPerPixel) / 8); }
	PixelRef operator*() { return *(PixelPtr)(src->data + ((mtc.x + mtc.y * src->size.x) * src->bitsPerPixel) / 8); }
};

/*
type-specific operation classes
*/
template<typename T>
class ImageType : public Image {
	typedef Image super;
protected:
	ImageType(const vec2i &size_, unsigned char *data_) : super(size_, sizeof(T)*8, data_) {}
public:
	ImageType() : super(vec2i(), sizeof(T)*8) {}
	ImageType(const vec2i &size_) : super(size_, sizeof(T)*8) {}

	REPEAT_WITH_CONST( T &pixel(vec2i tc, bool modulo = false) , {
		assert(data);
		assert(bitsPerPixel && (bitsPerPixel & 7) == 0);
		if (modulo) {
			for (int i = 0; i < tc.dim(); i++) {
				tc[i] %= size[i];
				if (tc[i] < 0) tc[i] += size[i];
			}
		} else {
			for (int i = 0; i < tc.dim(); i++) {
				if (tc[i] < 0) tc[i] = 0;
				if (tc[i] >= size[i]) tc[i] = size[i]-1;
			}
		}
		int bytesPerPixel = bitsPerPixel >> 3;
		int index = (tc.x + tc.y * size.x) * bytesPerPixel;
		return *(T*)(data + index);
	})
	REPEAT_WITH_CONST( T &pixel(int x, int y, bool modulo = false), {
		return pixel(vec2i(x,y),modulo);
	})

	typedef ImagePixelIterator<Image*, T &, T *> iterator;
	typedef ImagePixelIterator<const Image*, const T &, const T *> const_iterator;
	
	const_iterator begin() const { return const_iterator(this, 0); }
	const_iterator end() const { return const_iterator(this, size.x*size.y); }
	iterator begin() { return iterator(this, 0); }
	iterator end() { return iterator(this, size.x*size.y); }

	/*
	I would use a cast operator, but that'd imply stack allocation...
	and I would switch to heap allocation for the base data (think stl strings)
	but then that'd mean something about ref counting...
	maybe later...
	*/
	template<typename U>
	ImageType<U> *convert() const {
		ImageType<U> *dst = new ImageType<U>(size);
		//can't use an iterator cuz of the templates and member type...
		vec2i tc;
		for (tc.y = 0; tc.y < size.y; tc.y++) {
			for (tc.x = 0; tc.x < size.x; tc.x++) {
				dst->pixel(tc) = (U)pixel(tc);
			}
		}
		return dst;
	}
	
	ImageType *cloneEmpty() const { 
		return new ImageType(size);
	}
	
	ImageType *clone() const {
		unsigned char *newData = new unsigned char[getDataSize()];
		memcpy(newData, data, getDataSize());
		return new ImageType(size, newData);
	}
	
	ImageType *flip() const {
		assert(data);
		unsigned char *flipData = new unsigned char [getDataSize()];
		assert(flipData);
		int rowSize = getRowSize();
		for (int y = 0; y < size.y; y++) {
			memcpy(flipData + (size.y - y - 1) * rowSize, data + y * rowSize, rowSize);
		}
		return new ImageType(size, bitsPerPixel, flipData);
	}

	ImageType *resample(const vec2i &newSize) const {
		assert(!(bitsPerPixel&7));
		int bytesPerPixel = bitsPerPixel >> 3;
		unsigned char *newData = new unsigned char[newSize.x * newSize.y * 3];	//assert bpp == 24
		unsigned char *dst = newData;
		for (int j = 0; j < newSize.y; j++) {
			for (int i = 0; i < newSize.x; i++, dst+=bytesPerPixel) {
				vec2d f( (double)i / (double)(newSize.x-1), (double)j / (double)(newSize.y-1) );	//fractional coordinates
				vec2i r = (vec2i)vecScaleElem(f, (vec2d)(size-1));
				unsigned char *src = data + bytesPerPixel * (r.x + r.y * size.x);
				memcpy(dst, src, bytesPerPixel);
			}
		}
		return new ImageType(newSize, bytesPerPixel, newData);
	}

};

typedef ImageType<vec3uc> RGBImage;

/**
 * Abstract image loading interface.  create and register as you go.
 * Intertwined within it - through static methods - is access to the loader system
 */
class ImageLoader {
	
	//the image loader chain:
	ImageLoader *next;

public:

	ImageLoader *getNext();

	//attach yerself to the chain!
	void init();
	//detach on dtor? only if ever you expect to declare one non-global...

	//returns the image in the file.
	//if fails then it'll return NULL
	//at any time it returns an image, you should be able to assert that 'data' exists
	virtual Image *load(const char *filename, const char *fileExt) = 0;

	//return the name of the loader, for identification
	virtual const char *name() = 0;
};

/**
 * holding info and access to all loaders
 * currently through the static members of the ImageLoader interface
 */
class ImageSystem {
public:
	//the base of the image loader chain
	//todo - only allow this to be accessible by new image loader classes
	ImageLoader *baseLoader;

	ImageSystem() : baseLoader(NULL) {}

	//init everything - you gotta manually store all links to the loaders here
	//or else we may compile them in the wrong order and some be lost
	void init();

	//access to loading an image from our chain:
	Image *load(const char *filename);
};
extern ImageSystem imageSystem;

#endif //BASE_IMAGE_H