#ifndef BASE_PROFILER_H
#define BASE_PROFILER_H

#ifdef USE_PROFILER

#include <map>
#include <vector>
#include "system/singletons.h"
#include "defs/defs.h"
#include "defs/singleton.h"
#include "math/vec.h"	//specificially for #undef min and max

class ProfileEntry {
public:
	double min;
	double max;
	double avg;
	int counter;
	std::map<int, int> lastLocations;
	
	ProfileEntry() : min(0), max(0), avg(0), counter(0) {}
	
	void record(double delta, int lastLocationUid) {
		min = MIN(min, delta);
		max = MAX(max, delta);
		avg = (avg * (double)counter + delta) / (double)(counter + 1);
		counter++;
		lastLocations[lastLocationUid]++;
	}
};

class ProfileLocation {
public:
	const char *file;
	int line;
	const char *func;
	ProfileLocation(const char *file_, int line_, const char *func_)
	: file(file_), line(line_), func(func_) {}
	int uid() const { return (long int)file ^ line ^ (long int)func; }
	bool operator<(const ProfileLocation &p) const { return uid() < p.uid(); }
};

class Profiler;
class ProfileSystem {
protected:
	friend class Profiler;
	std::map<ProfileLocation, ProfileEntry> entries;
	std::vector<ProfileLocation> callStack;
public:
	~ProfileSystem() { report(); }
	void report();
};
extern Singleton<ProfileSystem> profileSys;

class Profiler {
	const char *file;
	int line;
	const char *func;
	double startTime;
	double endTime;
public:
	Profiler(const char *file_, int line_, const char *func_)
	: file(file_), line(line_), func(func_), startTime(timer.getTime()), endTime(-1) {
		profileSys->callStack.push_back(ProfileLocation(file, line, func));
		
	}
	~Profiler() {
		profileSys->callStack.pop_back();
		ProfileEntry *p = &(profileSys->entries[ProfileLocation(file, line, func)]);
		int lastuid = 0;
		if (profileSys->callStack.size()) lastuid = profileSys->callStack.rbegin()->uid();
		p->record(timer.getTime() - startTime, lastuid);
	}
};

//linking projects are to...
//(a) populate their profiled functions/code blocks with PROFILE() at the beginning 
//(b) define the USE_PROFILER macro when they want profile information generated
#define PROFILE()	Profiler profiler(__FILE__, __LINE__, __FUNCTION__)
#else
#define PROFILE()
#endif

#endif	//BASE_PROFILER_H
